Time = {};

-- Converts a number of seconds to a display string such as "1d 2h 30m 45s"
-- If 'canonical' is true, then all components will be returned; if false, 
-- approximations for long delays will be returned, such as "145 days".
-- If the argument is <= 0, returns "Now" (unless 'showNegativeTimes' is true).
function Time.GetDelayStr(delay, canonical, showNegativeTimes)
    if (type(delay) ~= "number") then
        error("delay must be a number", 2);
    end
    
    -- Format the time display string
    local text;
    delay = math.ceil(delay);
    local s = math.floor(delay % 60);
    local m = math.floor((delay % 3600) / 60);
    local h = math.floor((delay % 86400) / 3600);
    local d = math.floor(delay / 86400);
    if (canonical) then
        text = L:GetText("/Time/TimeDisplay/dhms");
        text = string.gsub(string.gsub(string.gsub(string.gsub(text, "<d>", d), "<h>", h), "<m>", m), "<s>", s);
    else
        if (delay <= 0.5) then
            if (showNegativeTimes and (delay <= -0.5)) then
                local delayStr = Time.GetDelayStr(-delay, canonical, false);
                text = L:GetText("/Time/TimeDisplay/NegativeTime");
                text = string.gsub(text, "<delay>", delayStr);
            else
                text = L:GetText("/Time/TimeDisplay/Now");
            end
        elseif (delay < 60) then
            text = L:GetText("/Time/TimeDisplay/s");
            text = string.gsub(text, "<s>", s);
        elseif (delay < 10 * 60) then
            text = L:GetText("/Time/TimeDisplay/ms");
            text = string.gsub(string.gsub(text, "<m>", m), "<s>", s);
        elseif (delay < 60 * 60) then
            text = L:GetText("/Time/TimeDisplay/m");
            text = string.gsub(text, "<m>", m);
        elseif (delay < 24 * 60 * 60) then
            text = L:GetText("/Time/TimeDisplay/hm");
            text = string.gsub(string.gsub(text, "<h>", h), "<m>", m);
        elseif (delay < 36 * 60 * 60) then
            text = L:GetText("/Time/TimeDisplay/dhm");
            text = string.gsub(string.gsub(string.gsub(text, "<d>", d), "<h>", h), "<m>", m);
        elseif (delay < 3 * 24 * 60 * 60) then
            text = L:GetText("/Time/TimeDisplay/dh");
            text = string.gsub(string.gsub(text, "<d>", d), "<h>", h);
        else
            if (h >= 12) then
                return Time.GetDelayStr(delay + 12 * 60 * 60);
            else
                text = L:GetText("/Time/TimeDisplay/d");
                text = string.gsub(text, "<d>", d);
            end
        end
    end
    return text;
end

-- Converts a delay string such as "1d 2h 30m 45s" into a number of seconds
function Time.GetNumericDelay(str)
    local delay = 0;
    local regexp = L:GetText("/Time/TimeInput/Regexp");
    local size = L:GetText("/Time/TimeInput/Components");
    for a, b in string.gmatch(str, regexp) do
        delay = delay + a * size[b];
    end
    return delay;
end

-- Converts hours, minutes into "12:34" or "12h34"
function Time.GetTimeOfDayStr(hour, minute)
    hour = string.format("%02d", tonumber(hour));
    minute = string.format("%02d", tonumber(minute));
    local str = L:GetText("/Time/TimeOfDay");
    str = string.gsub(str, "<h>", hour);
    str = string.gsub(str, "<m>", minute);
    return str;
end

-- Extracts hours, minutes from "12:34" or "12h34"
function Time.GetNumericTimeOfDay(str)
    local regexp = L:GetText("/Time/TimeOfDayRegexp");
    hour, minute = string.match(str, regexp);
    return hour, minute;
end

-- Accepts a time of day (specified as hours, minutes, seconds) and finds the next
-- GameTime at which that time will occur.
-- 'prevTime' optionally indicates the time after which to start searching (default = current game time)
function Time.GetNextOccurrenceOfTimeOfDay(hour, minute, second, prevTime)
    local gameTime = Turbine.Engine.GetGameTime();
    local dateInfo = Turbine.Engine.GetDate();
    local nextTime = gameTime + (hour - dateInfo.Hour) * 60 * 60;
    nextTime = nextTime + (minute - dateInfo.Minute) * 60;
    nextTime = nextTime + (second - dateInfo.Second);
    prevTime = prevTime or gameTime;
    if (nextTime <= prevTime) then
        local dayLength = 24 * 60 * 60;
        local moreDaysNeeded = math.ceil((prevTime - nextTime) / dayLength);
        nextTime = nextTime + (moreDaysNeeded * dayLength);
    end
    return nextTime;
end

-- Accepts a time of week (specified as dayofWeek, hours, minutes, seconds) and finds the next
-- GameTime at which that time will occur.
-- dayOfWeek 1...7 => Sunday...Saturday
-- 'prevTime' optionally indicates the time after which to start searching (default = current game time)
function Time.GetNextOccurrenceOfTimeOfWeek(dayOfWeek, hour, minute, second, prevTime)
    local gameTime = Turbine.Engine.GetGameTime();
    local dateInfo = Turbine.Engine.GetDate();
    local nextTime = gameTime + (dayOfWeek - dateInfo.DayOfWeek) * 24 * 60 * 60;
    nextTime = nextTime + (hour - dateInfo.Hour) * 60 * 60;
    nextTime = nextTime + (minute - dateInfo.Minute) * 60;
    nextTime = nextTime + (second - dateInfo.Second);
    prevTime = prevTime or gameTime;
    if (nextTime <= prevTime) then
        local weekLength = 7 * 24 * 60 * 60;
        local moreWeeksNeeded = math.ceil((prevTime - nextTime) / weekLength);
        nextTime = nextTime + (moreWeeksNeeded * weekLength);
    end
    return nextTime;
end

-- 'timeSpec' is a semicolon-separated list of day:time pairs, where
--   'day' can be 1...7 => Sunday...Saturday, or * => every day
--   'time' can be a fractional number
-- 'prevTime' optionally indicates the time after which to start searching (default = current game time)
-- example: "2:3;*:6;7:21.5" => Monday at 3:00 AM, any day at 6:00AM, Saturday at 9:30 PM
function Time.GetNextOccurrence(timeSpec, prevTime)
    local nextTime = math.huge;
    for day, hour in string.gmatch(timeSpec, "([1-7*]):([0-9.,]+)") do
        hour, minute = math.modf(hour);
        minute = minute * 60;
        if (day == "*") then
            nextTime = math.min(nextTime, Time.GetNextOccurrenceOfTimeOfDay(hour, minute, 0, prevTime));
        else
            nextTime = math.min(nextTime, Time.GetNextOccurrenceOfTimeOfWeek(day, hour, minute, 0, prevTime));
        end
    end
    return nextTime;
end

-- Used for converting a timeSpec from server time to local "game time"
function Time.AdjustTimeSpec(timeSpec, offsetHours)
    local newSpec = {};
    for day, hour in string.gmatch(timeSpec, "([1-7*]):([0-9.,]+)") do
        hour = tonumber(hour) + offsetHours;
        local dayOffset = 0;
        if (hour < 0) then
            hour = hour + 24;
            dayOffset = -1;
        elseif (hour >= 24) then
            hour = hour - 24;
            dayOffset = 1;
        end
        if (tonumber(day)) then
            day = day + dayOffset;
            if (day < 1) then
                day = 7;
            elseif (day > 7) then
                day = 1;
            end
        end
        table.insert(newSpec, day .. ":" .. tostring(hour));
    end
    return table.concat(newSpec, ";");
end

-- Gets rid of duplicate info in a timespec
function Time.GetCanonicalTimeSpec(timeSpec)
    local newSpec = {};

    -- Expand
    local schedule = { {}, {}, {}, {}, {}, {}, {} };
    local hours = {};
    for day, hour in string.gmatch(timeSpec, "([1-7*]):([0-9.,]+)") do
        hours[hour] = true;
        if (day == "*") then
            for d = 1, 7 do
                schedule[d][hour] = true;
            end
        else
            schedule[tonumber(day)][hour] = true;
        end
    end
    
    -- Contract
    local function isDaily(hour)
        for d = 1, 7 do
            if (not schedule[d][hour]) then
                return false;
            end
        end
        return true;
    end
    for hour in sorted_keys(hours) do
        if (isDaily(hour)) then
            table.insert(newSpec, "*:" .. hour);
        end
    end
    for d = 1, 7 do
        for hour in sorted_keys(schedule[d]) do
            if (not isDaily(hour)) then
                table.insert(newSpec, tostring(d) .. ":" .. hour);
            end
        end
    end
    
    return table.concat(newSpec, ";");
end
